"""
Python3 script to download ERA5 reanalysis data.
Currently the whole dataset for the globe is downloaded!

This version is dedicated to work on the super-computing
cluster "sophia" at DTU

Small modifications might be needed when 
using this script on another computer. 

DEPENDENCIES: 1) Account for ECMWF Downlaad (Copernicus)
              2) cdsapi module for python3 placed in same folder as this script 
              3) python3 with modules: os, cdo, shutil, datetime
              3) path "orig" to place the ERA5-Data

USAGE: python3 get_era5.py 2001-01-01 2001-01-31

2017-07-06 - V0 - Basics
2017-07-11 - V1 - Splitting working
2018-10-29 - V2 - Timeout added 
2019-11-05 - V3 - DTU version using cdsapi 
                  (based in part on Martin Doerenkaemper, IWES)
2020-04-16 - Modifications needed for sophia

TODO: 

author(s): Andrea Hahmann and Martin Doerenkaemper

(c) DTU
"""

import os
import re
import sys
import glob
import shutil
import datetime
import cdsapi
import subprocess
import multiprocessing
from cdo import *
c = cdsapi.Client()


def downdat(day,plfn,sffn):

    daystr = day.strftime("%Y-%m-%d")
    dayfns = day.strftime("%Y%m%d")
    year = day.strftime("%Y%m%d")
    
    c.retrieve(
        "reanalysis-era5-single-levels",
        {
            "product_type": 'reanalysis',
            "variable" : ["z","10u","10v","2t","2d","lsm","sp","msl","skt","ci","sst","rsn","sd","stl1","stl2","stl3","stl4","swvl1","swvl2","swvl3","swvl4"],
            "date"     : daystr+"",
            "time"     : ["00","06","12","18"],
            'format':'grib'
        },
        sffn+dayfns+".grb")

    c.retrieve(
        'reanalysis-era5-pressure-levels',
        {
            'product_type':'reanalysis',
            'variable':["z","t","u","v","r"],
            'pressure_level':[
                '10','20','30','50','70','100',
                '125','150','175','200','225','250',
                '300','350','400','450','500','550',
                '600','650','700','750','775','800',
                '825','850','875','900','925','950',
                '975','1000'
            ],
            "date"     : daystr+"",
            "time"     : ["00","06","12","18"],
            'format':'grib'
        },
        plfn+dayfns+".grb")

def get_and_split(sdate,edate,ipath,opath,plfn="ERA5_pl_",sffn="ERA5_sfc_"):
    """
    Function to call the download function for each day, 
    split the daily files to hourlies afterwards
    and move them to their final correct position in the file system       
    """
    adate = sdate
    while adate <= edate:
        print( "+++++++++++++++++ ERA-5 +++++++++++++++++++")
        print( "++++++++ Download data for "+adate.strftime("%Y-%m-%d"))
        # download_process=multiprocessing.Process(target=downdat,args=(adate,plfn,sffn))
        # download_process.start()
        # download_process.join(timeout=60*60*12) # Cancel Download if it takes more than 12 hours
        # if download_process.is_alive():
        #     download_process.terminate()
        #     download_process.join()
        #sstat = "Done"
        yr = str(adate.year).zfill(4)
        mn = str(adate.month).zfill(2)
        fpath = opath + yr + "/"
        ## ERA5_19930717_0000.grb
        if not os.path.exists(fpath):
           os.makedirs(fpath)
        ffile = "ERA5_"+adate.strftime("%Y%m%d") + "_0000.grb"
        if not os.path.isfile(fpath + ffile):
           print(fpath + ffile,"not there")
           sstat = downdat(adate,plfn,sffn)
           pfile = plfn + adate.strftime("%Y%m%d") + ".grb"
           sfile = sffn + adate.strftime("%Y%m%d") + ".grb"
           ofile = "ERA5_"+adate.strftime("%Y%m%d") + ".grb"
           print( "++++++++ Split data for "+adate.strftime("%Y-%m-%d"))
           cdo = Cdo()
           cdo.sellevel('1000/100000', input=pfile, output='ERA5_'+adate.strftime("%Y-%m-%d")+'_test2.grb')
           cdo.merge(input='ERA5_'+adate.strftime("%Y-%m-%d")+'_test2.grb '+sfile, output=ofile)
           cdo.splithour(input=ofile,output=ofile.split(".grb")[0]+"_")
           fall = glob.glob("ERA5_"+adate.strftime("%Y%m")+"*.grb")
           for fn1 in fall:
               if re.match('ERA5_\d{4}\d{2}\d{2}_\d{2}.grb',fn1):
                   fnn2 = fn1.split(".grb")[0] + "00.grb"
                   fnn2 = fpath + fnn2
                   shutil.move(fn1,fnn2)
           os.remove("ERA5_"+adate.strftime("%Y-%m-%d")+"_test2.grb")
           os.remove(ofile)
           os.remove(pfile)
           os.remove(sfile)

        adate = adate + datetime.timedelta(days=1)
        print( "+++++++++++++++++++++++++++++++++++++++++++")

    return(True)

def main():
    pass

    # Define Start and End Date for Download
    sdate = datetime.datetime.strptime(sys.argv[1],"%Y-%m-%d")
    edate = datetime.datetime.strptime(sys.argv[2],"%Y-%m-%d")

    # Define Path for storing the final data
    bpath = "/groups/reanalyses/"
    ipath = bpath
    opath = bpath + "era5/bdy/"

    # Run Download and Split Functions
    sstat = get_and_split(sdate,edate,ipath,opath)

    if sstat:
        print("All Done!")
    else:
        print("Something went wrong!")
        
        
if __name__ == '__main__':
    main()

  
