Module module_wf_ewp

!!! By P. Volker, DTU Windenergy
!!! Explicit Wake Parametrisation (EWP) V1.0
!!! The Explicit Wake Parametrisation: Development, implementation in a mesoscale
!!! model and evaluation against measurements, Geosci. Model Dev.

!!! This module is suitable for the Serial, Shared or Distributed Memory option
!!! When activated in the namelist, this module is called from the pbl_physics routine:

!!! (1) Initialisation (TURBINE_INIT): 
!!!     It reads the turbine positions and turbine type from the in the namelist 
!!!     specified file
!!! (2) Wake calculation (TURBINE_SELF_PRESERV):
!!!     calculates the deceleration for every specified turbine and applis it to 
!!!     the WRF grid-cells

  USE module_configure, ONLY : grid_config_rec_type

  USE module_model_constants, only: piconst
       
  USE module_driver_constants, ONLY : max_domains

  USE module_state_description, ONLY : EWPSCHEME

  IMPLICIT NONE


  INTEGER, PARAMETER :: WIND_TURBINES_OFF      = 0 ,&
       &                WIND_TURBINES_IDEAL    = 1 ,&
       &                WIND_TURBINES_REAL     = 2
  
  REAL  , PARAMETER   ::                                     &
       z0         = 0.0 ,       z1         = 1.0            ,&
       z2         = 2.0 ,       z3         = 3.0            ,&
       piGr       = piconst,    z1d2       = z1/z2          ,&    
       z2d3       = z2/z3,      z3d2       = z3/z2          ,&     
       eps        = 1e-3,       large      = 1e6            ,&  
       sqrt2      = 1.41421356, sqrt_pi    = 1.772453851    ,&
       undef      = -9999.,     KMmin      = 0.1            ,&      
       Rearth     = 6371000.,   D2R        = piGr/180.      ,&
       Betz       = 16./27.,    sqrt8      = 2.83           ,&
       rhoA       = 1.224


  INTEGER , PARAMETER ::                                     &
       n0          = 0 ,         n1        = 1              ,&
       n2          = 2 ,         n5        = 5              ,&
       error       = 999,        nCt       = 100


  CHARACTER(len=256) , PARAMETER ::   turb_type = 'Specified'

  CHARACTER(len=120)             :: farmname

  LOGICAL , EXTERNAL      :: wrf_dm_on_monitor

  !In V1.1 exchange the integer cluster_nr to the real Nturbines (number of turbines)
  !Every turbine has the properties:
  TYPE turbine_specs
     REAL                               :: hubheight  ! height (H) of the turbine hub (m)
     REAL                               :: Nturbs     ! Number of specified turbines at a given location 
!     INTEGER                            :: nCt        ! number of u,Ct,Cp enteries in the files
     REAL   , DIMENSION(:), ALLOCATABLE :: Ut         ! Thrust curve U
     REAL   , DIMENSION(:), ALLOCATABLE :: Ct         ! Thrust curve Ct
     REAL   , DIMENSION(:), ALLOCATABLE :: Cp         ! Thrust curve Cp
     INTEGER                            :: turb_nr    ! which turbines are in cell(n)
     REAL                               :: Ct_stat
     REAL                               :: diameter    
     REAL                               :: cutinspeed  
     REAL                               :: cutoutspeed 
     REAL                               :: wakes
  END TYPE turbine_specs

  TYPE COORDINATE
     REAL                               :: i, j ! x and y coords of turbines
  END TYPE COORDINATE

  !Cells are used to apply the accumulative Tc
!  TYPE cell_specs
!     INTEGER                            :: i, j        ! x and y coords of turbines
!     INTEGER                            :: turb_p_cell ! nr of turbines in cell(n)
!     INTEGER, DIMENSION(:), ALLOCATABLE :: turb_nr     ! which turbines are in cell(n)
!  END TYPE cell_specs


  INTEGER                               :: nturbines
!  INTEGER   , DIMENSION(max_domains)    :: ncells

  TYPE(turbine_specs), DIMENSION(:)  , ALLOCATABLE :: turbines


  INTEGER, DIMENSION(:,:), ALLOCATABLE :: posi,posj

!  TYPE(cell_specs)   , DIMENSION(:,:), ALLOCATABLE :: cells
      
  !THIS IS NEEDED ONLY FOR THE ALLOCATION OF THE SLAVE CORES
!  INTEGER   , DIMENSION(:,:), ALLOCATABLE :: TPC_CNT !DIMENSION NCELLS

  PUBLIC  TURBINE_EWP
  PUBLIC  INIT_EWP

  PRIVATE WAKE         ,&
       &  PREP_TURB    ,&
!       &  PREP_CELL    ,&
       &  GET_SPEC     ,&
       &  ABORT

 CONTAINS
    
  ! Main routines:
  ! - TURBINE_SELF_PRESERV
  ! - WAKE
  ! Utility routines:
  ! - TURBINE_INIT
  ! - PREP_TURB
  ! - PREP_CELL
  ! - GET_WIND_DIR
  ! - ABORT

!!! #########################################################################
!!! ######################### BEGIN PUBLIC ROUTINES #########################
!!! #########################################################################

!!! #########################################################################
!!!                                  (1)                                  !!!
!!! #########################################################################    
  SUBROUTINE INIT_EWP(dx,id,config_flags,                            &
         &            lat,lon,                                       &
         &            IDS,IDE,JDS,JDE,KDS,KDE,                       &
         &            IMS,IME,JMS,JME,KMS,KME,                       &
         &            ITS,ITE,JTS,JTE,KTS,KTE)

    INTEGER,INTENT(IN)           :: IDS,IDE,JDS,JDE,KDS,KDE,         &
         &                          IMS,IME,JMS,JME,KMS,KME,         &
         &                          ITS,ITE,JTS,JTE,KTS,KTE


    REAL, DIMENSION(ims:ime,jms:jme), INTENT(IN) :: lon,lat
    REAL,                             INTENT(IN) :: dx
    INTEGER,                          INTENT(IN) :: id
   
    CHARACTER(len=256)                           :: message
    INTEGER                                      :: ITF,JTF,KTF,i,j,k
    LOGICAL                                      :: INIT

    TYPE (grid_config_rec_type) :: config_flags

    INIT = .FALSE.

    JTF=MIN0(JTE,JDE-n1)
    KTF=MIN0(KTE,KDE-n1)
    ITF=MIN0(ITE,IDE-n1)

    write(message,*)'INSIDE TURBINE',wrf_dm_on_monitor(),id
    CALL wrf_debug(00,message)

!    IF(.not.ALLOCATED(turbines))THEN

    IF (wrf_dm_on_monitor() ) THEN
 
       WRITE(message,*)'INIT_EWP',id
       CALL wrf_debug(000,message)

       CALL PREP_TURB(dx=dx,id=id                                    ,&
            &         known_lat=lat(its,jts),known_lon=lon(its,jts)  ,&
            &         ITS=ITS,JTS=JTS,                                &
            &         KTS=KTS,KTF=KTF,IMS=IMS,IME=IME,JMS=JMS,JME=JME,&
            &         config_flags=config_flags)

    END IF

    CALL wrf_dm_bcast_integer(nturbines ,1)
    !CALL wrf_dm_bcast_integer(nCt       ,1)
 
    IF(.NOT.wrf_dm_on_monitor() )THEN

       IF(.NOT.ALLOCATED(turbines) )THEN

 write(message,*)'ALLOCATE_EWP',id
 CALL wrf_debug(000,message)

          ALLOCATE(turbines(nturbines))
          DO i = 1, nturbines

             IF(.NOT.ALLOCATED(turbines(i)%Ut))&
                     ALLOCATE( turbines(i)%Ut(nCt))
             IF(.NOT.ALLOCATED(turbines(i)%Ct))&
                     ALLOCATE( turbines(i)%Ct(nCt))
             IF(.NOT.ALLOCATED(turbines(i)%Cp))&
                     ALLOCATE( turbines(i)%Cp(nCt))

          END DO

          DO i = 1, nturbines
             turbines(i)%Ut(:) = z0
             turbines(i)%Ct(:) = z0
             turbines(i)%Cp(:) = z0
          END DO

          INIT = .TRUE.
       END IF

       IF(.NOT.ALLOCATED(posi) )THEN
          ALLOCATE( posi(nturbines,max_domains))
          posi(:,:) = n0
       END IF

       IF(.NOT.ALLOCATED(posj) )THEN
          ALLOCATE( posj(nturbines,max_domains))
          posj(:,:) = n0
       END IF

    END IF



       CALL wrf_dm_bcast_integer(posi(:,:),nturbines*max_domains) 
       CALL wrf_dm_bcast_integer(posj(:,:),nturbines*max_domains) 

!       IF(INIT)THEN
          DO i = 1, nturbines 
             CALL wrf_dm_bcast_real(   turbines(i)%hubheight  ,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%Nturbs     ,1) 
!             CALL wrf_dm_bcast_integer(turbines(i)%nCt        ,1)
             DO k = 1, nCt
                CALL wrf_dm_bcast_real(turbines(i)%Ct(k)      ,1) 
                CALL wrf_dm_bcast_real(turbines(i)%Cp(k)      ,1) 
                CALL wrf_dm_bcast_real(turbines(i)%Ut(k)      ,1)
             END DO
             CALL wrf_dm_bcast_integer(turbines(i)%turb_nr    ,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%Ct_stat    ,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%diameter   ,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%cutinspeed ,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%cutoutspeed,1) 
             CALL wrf_dm_bcast_real(   turbines(i)%wakes      ,1) 
          END DO
!       END IF



       write(message,*)'Number of turbines:',nturbines
       CALL wrf_debug(200,message)
!       CALL wrf_message(message)





  END SUBROUTINE INIT_EWP
!!! #########################################################################
!!! #########################################################################



!!!
!!!
!!!
!!! #########################################################################
!!!                                  (2)                                  !!!
!!! #########################################################################   
  SUBROUTINE TURBINE_EWP(id,windfarm_opt,u,v,dudt,dvdt,dx,dz,z,hgt,&
       &             rho,R0frac,k_m,power,                         &
       &             IDS,IDE,JDS,JDE,KDS,KDE,                      &
       &             IMS,IME,JMS,JME,KMS,KME,                      &
       &             ITS,ITE,JTS,JTE,KTS,KTE)
 
    
    IMPLICIT NONE

    INTEGER,INTENT(IN)        :: IDS,IDE,JDS,JDE,KDS,KDE,              &
         &                       IMS,IME,JMS,JME,KMS,KME,              &
         &                       ITS,ITE,JTS,JTE,KTS,KTE,id,windfarm_opt
    
    REAL    ,DIMENSION(ims:ime,kms:kme,jms:jme),INTENT(INOUT):: k_m
    REAL    ,DIMENSION(ims:ime,jms:jme),INTENT(INOUT):: Power
    REAL    ,DIMENSION(ims:ime,jms:jme),INTENT(IN)   :: hgt
    REAL    ,DIMENSION(ims:ime,kms:kme,jms:jme),INTENT(INOUT):: dudt,dvdt  
    REAL    ,DIMENSION(ims:ime,kms:kme,jms:jme),INTENT(IN   ):: U,V,z,dz,rho
    REAL    ,                                   INTENT(IN   ):: dx,R0frac

    REAL    ,DIMENSION(        kms:kme        )              :: Tc,phi

    CHARACTER(len=256)           :: message
    REAL                         :: R0,C_t,C_p,ell0,Depth,Area,L
    REAL                         :: hhl,D,fact,Nturb,Zmod
    REAL                         :: Km1,Km2,Km,U1,U2,U0

    INTEGER                      :: i,ii,j,jj,k,n,                     &
         &                          ITF,JTF,KTF,indi,ind,              &
         &                          iturb,icell


    !Actually module is outside of the tile loop, use only K component
    JTF=MIN0(JTE,JDE-n1)
    KTF=MIN0(KTE,KDE-n1)
    ITF=MIN0(ITE,IDE-n1)


    ! First time step only allocation, because we have to allocate only once
    ! then from the next timestep onwards, all tiles have to be used.
    !=========================================================================
    !This is just a way to do it. We go from cell to cell and at the end apply 
    !a thurst to it. 

    Power = z0

    TURBINE:DO iturb = n1,nturbines

       !i = NINT(turbines(iturb)%i(id))
       !j = NINT(turbines(iturb)%j(id)) 

       i = posi(iturb,id)
       j = posj(iturb,id)

       !write(message,*)'id:',id,' TURB:',iturb,'pos:',i,j
       !CALL wrf_debug(0,message)

       IF((ITS.LE.I.AND.I.LE.ITF).AND.(JTS.LE.J.AND.J.LE.JTF))THEN

          phi(:) = atan2(v(i,:,j),u(i,:,j)) + z2 * piGr
          phi(:) = mod(phi,z2 * piGr)

          !=====================================================================
          !We manipulate the i,j field where the turbines are located, with help 
          !of the reference field ii,jj. We assume therefore a local homogeneous 
          !mean velocity field (open terrain)
          Tc           = z0

          Nturb  = turbines(iturb)%Nturbs 
          hhl    = turbines(iturb)%hubheight
          R0     = z1d2 * turbines(iturb)%diameter

       !write(message,*)'id ',id,' hh R0 ',Nturb,hhl,R0
       !CALL wrf_debug(000,message)

          ell0  = R0frac * R0
          Depth = dx
          Area  = dx*dx
          L     = z1d2 * dx

          k = 1
          Zmod = Z(i,k,j) - HGT(i,j)
          do while(Zmod<hhl)
             k = k + 1
             Zmod = Z(i,k,j) - HGT(i,j)
             ind = k
          end do

          U1 = sqrt(U(i,ind-1,j)**n2 + V(i,ind-1,j)**n2)
          U2 = sqrt(U(i,ind  ,j)**n2 + V(i,ind  ,j)**n2)
   
          U0 =   U1 + (U2 - U1) * (log(hhl)                 - log(Z(i,ind-1,j)-hgt(i,j) ))&
             &                  / (log(Z(i,ind,j)-hgt(i,j)) - log(Z(i,ind-1,j)-hgt(i,j) ))
 
          Km1 = K_m(i,ind-1,j)
          Km2 = K_m(i,ind,j)
   
          Km =   Km1 + (Km2 - Km1) * (hhl - Z(i,ind-1,j)-hgt(i,j) )&
             &                     / (Z(i,ind,j) - Z(i,ind-1,j) )

!!!       =========  GET C_t & C_p  ===================================================
          IF(trim(turb_type).eq.'none')then !constant Ct assumed

             if(    U0.gt.turbines(iturb)%cutinspeed  .and.&
                  & U0.lt.turbines(iturb)%cutoutspeed )then
                C_t = 0.89
                C_p = Betz
             else
                C_t = turbines(iturb)%Ct_stat
                C_p = z0
             end if
             
          ELSE  ! C_t from specifications
             
             indi = n1
             !do while(U0.ge.turbines(iturb)%Ut(indi))
             do while(turbines(iturb)%Ut(indi).le.U0)
                !if(indi.ge.turbines(iturb)%nCt)goto 110
                !nCt has a standart length of 50. If it didn't find 
                !a maching wind speed, the wind speed will be above the latest entry
                if(indi.ge.nCt - n1)goto 110   
                indi = indi + n1
             end do
110          continue
             indi    = indi - n1

             !linear interpolation:
             if(    U0.gt.turbines(iturb)%cutinspeed .and.&
                  & U0.lt.turbines(iturb)%cutoutspeed )then
                C_t    =  (U0 - turbines(iturb)%Ut(indi)) &
                     & * (turbines(iturb)%Ct(indi+n1) - turbines(iturb)%Ct(indi))  &
                     & / max((turbines(iturb)%Ut(indi+n1) - turbines(iturb)%Ut(indi)),eps)  & 
                     & +  turbines(iturb)%Ct(indi)

                C_p    =  (U0 - turbines(iturb)%Ut(indi)) &
                     & * (turbines(iturb)%Cp(indi+n1) - turbines(iturb)%Cp(indi))  &
                     & / max((turbines(iturb)%Ut(indi+n1) - turbines(iturb)%Ut(indi)),eps)  & 
                     & +  turbines(iturb)%Cp(indi)
             else
                C_t = turbines(iturb)%Ct_stat
                C_p = z0
             end if

          END IF
!          C_p = 0.5
!          C_t = 0.9
!!!       =========  END GET C_t  ====================================================


       !write(message,*)'id ',id,' Ct Cp ',U0,C_t,C_p
       !CALL wrf_debug(000,message)

          fact = z2 * max(Km,KMmin) / max(U0,eps)

          !Calculate the thrust for one coordidinate
          CALL WAKE(ks=KTS              , ke=KTF               ,&
                &   U0m=U0              , hhl=hhl              ,&
                &    R0=R0              , Area=Area            ,&
                &    Z=Z(i,KTS:KTF,j)   , L=L                  ,&
                &    Nturb=Nturb        ,                       & 
                &    Tc=Tc              , fact=fact            ,&
                &    ell0=ell0          , Depth=Depth          ,&
                &    C_t=C_t            , hgt = hgt(i,j)        )

          !Eq. 14 and 15: add the thrust to the velocity tendency
          IF(turbines(iturb)%wakes.gt.z1d2.and.windfarm_opt.eq.EWPSCHEME)THEN
             dudt(i,:,j) = dudt(i,:,j) + cos(phi) * Tc
             dvdt(i,:,j) = dvdt(i,:,j) + sin(phi) * Tc
          END IF

          !Calculate the power production for one coordinate (the can have 
          !several turbines) use 1.23 similar to fitch
          Power(i,j) = Power(i,j) &
          &          + z1d2 * Nturb * 1.23 * C_p * piGr * R0**2 * U0**3
          !&          + z1d2 * Nturb * rhoA * C_p * piGr * R0**2 * U0**3

 write(message,&
  '(A4,1X,I4.4,A3,I3.3,1X,I3.3,A3,F9.1,1X,F5.2,1X,F5.3,1X,F7.1,1X,F4.1)')&
   'EWP:',iturb,' C:',i,j,' P:',Power(i,j),U0,C_p, piGr * R0**2,R0 !,C_t,U0
  CALL wrf_debug(199,message)

       END IF

    END DO TURBINE







  END SUBROUTINE TURBINE_EWP
!!! #########################################################################
!!! ######################### END PUBLIC ROUTINES ###########################
!!! #########################################################################






!!! #########################################################################
!!! ######################### BEGIN LOCAL ROUTINES ##########################
!!! #########################################################################
!!! 
!!! =========================================================================
!!!                   Model Grid                           Thrust (Tc)
!!!                                                       0
!!!  |                                            |       | |  
!!!  |                                            |       | |
!!!  ----------------------------------------------       | ----      
!!!  |                    |                       |       |     |
!!!  |                    |                       |       |     |
!!!  --------------       | //////     ------------       |     --
!!!  |               hub  |--||                   |       |       |        
!!!  |                    |  ||                   |       |       |
!!!  --------------       |  ||        ------------       |     --
!!!  |                    |  ||                   |       |     |
!!!  |                       ||                   |       |     |
!!!  --------------          ||        ------------       | ----
!!!  |                       ||                   |       | |
!!!  |                       ||                   |       | |
!!! =========================================================================
!!! Turbine wakes defined around hub height and the description 
!!! uses the wnd speed and turbulence diffusion at this height 
!!! to determine the grid-cell averaged wake extension in the 
!!! vertical direction. Therefore, the grid-cell averaged 
!!! velocity deficit is not necessarily bounded to the turbine's 
!!! blade. 

!!! #########################################################################
!!!              calculate the deceleration for every turbine
!!! #########################################################################
  SUBROUTINE WAKE(ks,ke,U0m,hhl,R0,Area,Z,L,Nturb,Tc  ,&
       &          fact,ell0,Depth,C_t,hgt              )
    
    IMPLICIT NONE
    
    REAL      ,DIMENSION(ks:ke)   ,INTENT(INOUT) :: Tc
    REAL                          ,INTENT(INOUT) :: hhl
    REAL      ,DIMENSION(ks:ke)   ,INTENT(IN   ) :: Z
    REAL                          ,INTENT(IN   ) :: L,R0,C_t,Area
    REAL                          ,INTENT(IN   ) :: ell0,Depth,hgt
    REAL                          ,INTENT(IN   ) :: fact,U0m,Nturb
    INTEGER                       ,INTENT(IN   ) :: ks,ke

    
    CHARACTER(len=256)          :: message 
    REAL                        :: Usmn,ellmn,Zhl
    REAL                        :: Xi
    INTEGER                     :: k,n


    !Eq. (12) (uses fact in denominater therefore looks slightly different)
    ellmn = z2d3 * ( (fact*L + ell0**2)**z3d2 - ell0**3) / (fact*L)

    !refers to Eq. (11) which was not a wind speed. Here it is:
    Usmn = C_t * sqrt_pi * R0**2 * U0m      & 
            &            / (sqrt8 * ellmn * Depth)

    !########################################################################
    !############### calculate the velocity deficit at each level ###########
    DO k=ks,ke-1

       !gaussian wake similar to Eq.6, but wti grid-cell av. ell, ellmn
       Xi = exp(-z1d2 * ( ( (Z(k)-hgt) - hhl) / ellmn)**n2)

       !Prepare Eq. 14 and 15
       Tc(k) = - Nturb * U0m * Usmn * Xi * Depth / Area

    END DO
    !########################################################################

        
  END SUBROUTINE WAKE 
!!! #########################################################################
!!! #########################################################################






!!! #########################################################################
!!! #########################################################################
!!! The input file where the locations of the individual turbines are defined, 
!!! will be read. The structure of the turbine will be alloc. and their
!!! elements be defined 
  SUBROUTINE PREP_TURB(dx,id,known_lat,known_lon                  ,&
                       ITS,JTS,KTS,KTF,IMS,IME,JMS,JME,config_flags)

    USE module_llxy

    IMPLICIT NONE


    INTEGER,INTENT(IN)       :: ITS,JTS,KTS,KTF,IMS,IME,JMS,JME,id
    REAL   ,INTENT(IN)       :: dx,known_lat,known_lon

    INTEGER                  :: ilon,ilat
    REAL                     :: diameter,cutinspeed,cutoutspeed,hubheight
    REAL                     :: Tlat,Tlon,frac,height,wakes
    REAL                     :: Lx,Ly,phi,Dist,nval,ts_rx,ts_ry
    INTEGER                  :: n,m,i,j,ii,jj
    INTEGER                  :: istat,dot,turbine_case
    CHARACTER(len=256)       :: heather,message
    CHARACTER(len=120)       :: inline
    CHARACTER(len=20)        :: device
    LOGICAL                  :: rfile,INIT
    INTEGER ,PARAMETER       :: ffarm = 19

    TYPE (PROJ_INFO) :: ts_proj
    TYPE (grid_config_rec_type) :: config_flags

    INIT = .FALSE.

!    write(message,*)'inside prep_turb'
!    CALL wrf_message(message)

    CALL nl_get_windturbines_spec( 1, farmname )

    dot = INDEX(TRIM(farmname),'.')    
    write(message,*)'rfile ',farmname
    CALL wrf_debug(200,message)
    !#### DEFINE IF FILE TO READ ############################################
    IF     (farmname(dot+n1:dot+n5).eq.'real' ) THEN  
       turbine_case = WIND_TURBINES_REAL
       rfile    = .true.
    ELSE IF(farmname(dot+n1:dot+n5).eq.'ideal') THEN  
       turbine_case = WIND_TURBINES_IDEAL
       rfile    = .true.
    ELSE                                          
       turbine_case = WIND_TURBINES_OFF
       rfile        = .false.
    END IF
    !########################################################################
    

    istat = n0
      write(message,*)'rfile ',rfile,turbine_case ; CALL wrf_debug(200,message)
    !########## READ FILE ###################################################
    readf:IF(rfile)THEN
       OPEN(file=TRIM(farmname),unit   = ffarm      ,&
            &                   FORM   = 'FORMATTED',&
            &                   STATUS = 'OLD'      ,&
            &                   IOSTAT = istat       )
       
       IF (istat .EQ. n0)THEN          
          ! get the number of turbines
          n = n0
          DO WHILE (.true.)
             READ(unit=ffarm,FMT='(A120)',END=30)inline
             write(message,*)inline; CALL wrf_debug(200,message)
             IF ( INDEX(inline,'!')   .EQ. n0 .and. &
                  INDEX(inline,'#')   .EQ. n0)&
                  n = n + n1
          ENDDO
30        CONTINUE

          nturbines = n    !number of turbines is the same for all domains
          !write(message,*)'id ',id,'nturbines ',nturbines; CALL wrf_message(message)
          !------  ALLOCATE TURBINES/MAST ------------------------------------
          IF(.not.ALLOCATED(turbines))THEN
             ALLOCATE(turbines(nturbines))
             turbines(:)%turb_nr    = n0 ; turbines(:)%hubheight   = z0
             turbines(:)%wakes      = z0
             turbines(:)%Nturbs     = n0 ; turbines(:)%cutinspeed  = z0 
             turbines(:)%diameter   = z0 ; turbines(:)%cutoutspeed = z0

             INIT = .TRUE.

          END IF

          IF(.not.ALLOCATED(posi))THEN
             ALLOCATE(posi(nturbines,max_domains))
             posi(:,:) = n0
          END IF

          IF(.not.ALLOCATED(posj))THEN
             ALLOCATE(posj(nturbines,max_domains))
             posj(:,:) = n0
          END IF
          !-------------------------------------------------------------------

          REWIND(ffarm)
                    
          heather   = 'none'
          device    = 'none'

          n    = n1
          m    = n1
          nval = z0 

          !--------------------------------------------------------------
          !Decide which kind of data we have
          !heather should start with ### followed by one space
          READ(ffarm,'(A120)',END=120)inline
          if(INDEX(inline,'latlon').gt.n0)then
             heather = 'latlon'
          else if(INDEX(inline,'lonlat').gt.n0)then
             heather = 'lonlat'
          else !Default same as WRF-WF:
             heather = 'lonlat'
          end if

          DO WHILE (.true.)

             READ(ffarm,'(A120)',END=120)inline

             !---------------------------------------------------------------
             INFO:IF ( n .le. nturbines.and.INDEX(inline,'!').eq.n0 &
                  &                    .and.INDEX(inline,'#').eq.n0)THEN     
                
                IF(trim(heather).eq.'latlon')THEN
                   READ(inline,*,ERR=130)nval,Tlat,Tlon,device,wakes
                ELSE
                   READ(inline,*,ERR=130)nval,Tlon,Tlat,device,wakes
                END IF
                
                !read the corresponding specifications
                IF(INIT)THEN 
                   CALL GET_SPEC(n=n,                 &
                   &   cutinspeed=cutinspeed,         &
                   &  cutoutspeed=cutoutspeed,        &
                   &    hubheight=hubheight,          &
                   &     diameter=diameter,           &
                   &       device=device              )

                   turbines(n)%turb_nr     = n
                   turbines(n)%Nturbs      = nval
                   turbines(n)%hubheight   = hubheight
                   turbines(n)%diameter    = diameter
                   turbines(n)%cutinspeed  = cutinspeed
                   turbines(n)%cutoutspeed = cutoutspeed
                   turbines(n)%wakes       = wakes
                END IF


                IF(turbine_case .eq. WIND_TURBINES_REAL)THEN !find IJ
                    !This can be updated for if we allow for lon lat positions

                   CALL map_init(ts_proj)

                   ! Mercator
                   IF (config_flags%map_proj == PROJ_MERC) THEN
                      CALL map_set(PROJ_MERC, ts_proj,     &
                         truelat1 = config_flags%truelat1, &
                         lat1     = known_lat,             &
                         lon1     = known_lon,             &
                         knowni   = REAL(its),             &
                         knownj   = REAL(jts),             &
                         dx       = dx)!config_flags%dx)

                   ! Lambert conformal
                   ELSE IF (config_flags%map_proj == PROJ_LC) THEN
                      CALL map_set(PROJ_LC, ts_proj,        &
                         truelat1 = config_flags%truelat1,  &
                         truelat2 = config_flags%truelat2,  &
                         stdlon   = config_flags%stand_lon, &
                         lat1     = known_lat,              &
                         lon1     = known_lon,              &
                         knowni   = REAL(its),              &
                         knownj   = REAL(jts),              &
                         dx       = dx)!config_flags%dx)
                   ! Polar stereographic
                   ELSE IF (config_flags%map_proj == PROJ_PS) THEN
                      CALL map_set(PROJ_PS, ts_proj,        &
                         truelat1 = config_flags%truelat1,  &
                         stdlon   = config_flags%stand_lon, &
                         lat1     = known_lat,              &
                         lon1     = known_lon,              &
                         knowni   = REAL(its),              &
                         knownj   = REAL(jts),              &
                         dx       = dx)!config_flags%dx)
                   END IF

                   CALL latlon_to_ij(ts_proj, Tlat, Tlon, ts_rx, ts_ry)

                   !matter of taste. With floor turbines remain in the 
                   !same cell and wih nint they can end up in the next one
                   ! floor(45.6)=45 and nint(45.6)=46, personally prefer floor

                   posi(n,id) = NINT(ts_rx)
                   posj(n,id) = NINT(ts_ry)

                   write(message,*)'postion',id,posi(n,id),posj(n,id) 
                   CALL wrf_debug(200,message)

                ELSE ! We have the position in distance (km) from origin i=1,j=1
                   
                   posi(n,id) = NINT(Tlon)
                   posj(n,id) = NINT(Tlat)

                END IF

                n = n + n1
                                
             ENDIF INFO
             
          ENDDO
          
120       CONTINUE ; CLOSE(ffarm) ;                 GOTO 150 
130       CONTINUE ; CLOSE(ffarm) ; istat = error ; GOTO 150 
          
       ENDIF
       
    END IF readf
150 CONTINUE

    IF (istat.eq.error)CALL ABORT('error reading file ')



  END SUBROUTINE PREP_TURB
!!! #########################################################################
!!! #########################################################################





!!! #########################################################################
!!! #########################################################################
!!! At the moment every turbine will have its own Ct curve, which is quiet 
!!! a wast of memory,. Give him a name and define one Ct curve where it 
!!! will be connected to. The actual situation has the advantage that in 
!!! principle every turbine could behave different
  SUBROUTINE GET_SPEC(n,cutinspeed,cutoutspeed,hubheight,diameter,device)
    
    IMPLICIT NONE
    
    INTEGER   ,INTENT(IN)    :: n !turbines(n)
    REAL      ,INTENT(INOUT) :: cutinspeed,cutoutspeed,hubheight,diameter
    REAL                     :: b,R2_d_R02

    character(len=20)        :: device
    character(len=256)       :: message
    CHARACTER(len=120)       :: specname,inline,spec
    INTEGER ,PARAMETER       :: fspec = 98
    REAL    ,PARAMETER       :: cpbetz=0.593 !16/27
    INTEGER                  :: istat,ih,is,i
    REAL                     :: val,cpmax
        
    istat = n0
    
    !Constant Ct
    IF(trim(device).eq.'none')then
       
       b = z2d3
       
       IF(.NOT.ALLOCATED(turbines(n)%Ct))THEN
          DO i=1,max_domains
             ALLOCATE(turbines(n)%Ct(n1))
             turbines(n)%Ct(:) = z0
          END DO
       END IF

       turbines(n)%Ct(n1)  =  b * (z2 - b) 
       turbines(n)%Ct_stat =  0.05
       
       cutinspeed       = cutinspeed
       cutoutspeed      = cutoutspeed
       hubheight        = hubheight
       diameter         = diameter
!       turbines(n)%nCt  = n1

    ELSE 
       
       write(specname,'(A,A8)')trim(device),'.turbine'
       
       OPEN(file=TRIM(specname),unit   = fspec      ,&
            &                   FORM   = 'FORMATTED',&
            &                   STATUS = 'OLD'      ,&
            &                   IOSTAT = istat       )
       
       IF (istat .EQ. n0)THEN
          
          ! get the number of turbines
          ih = n0  !heather
          is = n0  !specification
          DO WHILE (.true.)
             READ(fspec,'(A120)',END=30)inline
             IF ( INDEX(inline,'!')   .EQ. n0)THEN
                is = is + n1
             ELSE
                ih = ih + n1
             END IF
          ENDDO
30        CONTINUE
          
          !turbines(n)%nCt = is

          ALLOCATE(turbines(n)%Ut(nCt))
          ALLOCATE(turbines(n)%Ct(nCt))
          ALLOCATE(turbines(n)%Cp(nCt))
          
          turbines(n)%Ut(:) = z0
          turbines(n)%Ct(:) = z0
          turbines(n)%Cp(:) = z0

          !write(message,*)'Before Turbine ',n,turbines(n)%nCt,max_domains
          !CALL wrf_message(message)

          REWIND(fspec)

          
          DO i=n1,ih
             READ(fspec,*)spec,val
             SELECT CASE(trim(spec))
             CASE('!cutinspeed' );cutinspeed          = val
             CASE('!cutoutspeed');cutoutspeed         = val
             CASE('!hubheight'  );hubheight           = val
             CASE('!diameter'   );diameter            = val
             CASE('!Ct_stat'    );turbines(n)%Ct_stat = val
             END SELECT
          END DO
          
          !DO i=n1,nCt
          i = n1
          DO WHILE (.true.)
             READ(fspec,*,END=40)&
                   turbines(n)%Ut(i),turbines(n)%Ct(i),turbines(n)%Cp(i)
              i = i + n1
          END DO
 40       CONTINUE
          
          CLOSE(fspec)

          cpmax = maxval(turbines(n)%Cp(:))

          if(cpmax.gt.cpbetz) turbines(n)%Cp(:) =      turbines(n)%Cp(:) / &
             (z1d2 * rhoA *piGr * (z1d2*diameter)**2 * turbines(n)%Ut(:)**3)
          
       ELSE
          
          write(message,*) 'Failed to open turbine file:', trim(specname)     
          CALL wrf_debug(200,message)
          
          write(message,*) 'Please ensure turbine file name has less than 20 characters.'  
          CALL wrf_debug(200,message)

          write(specname,'(A16,A,A15)')&
               'No turbine file ',trim(specname),' could be found. Please ensure turbine file name has less than 20 characters.'
          
          CALL ABORT(trim(specname))
          
       END IF
       
    END IF

  END SUBROUTINE GET_SPEC
!!! #########################################################################
!!! #########################################################################






!!! #########################################################################
!!! #########################################################################
!!! Define the flow direction, then fill in all elements of the structure cells 
!  SUBROUTINE PREP_CELL(kms,kme,ims,ime,jms,jme,JTS,JTF,ITS,ITF,KTS,KTF,dx,id)
    
!    IMPLICIT NONE
    
!    REAL                              ,INTENT(IN)::dx
!    INTEGER                           ,INTENT(IN)::ims,ime,its,itf
!    INTEGER                           ,INTENT(IN)::jms,jme,jts,jtf
!    INTEGER                           ,INTENT(IN)::kms,kme,kts,ktf,id

!    INTEGER                         :: iturb,cnt,indi,indo,k
!    INTEGER                         :: i,ii,j,jj,n,nn,o,si,fi,so,fo
!    INTEGER                         :: mini,maxi,minj,maxj,hcnt
!    LOGICAL                         :: find


!    mini = large
!    minj = large
!    maxi = z0
!    maxj = z0

!    ! Get the mean wind direction
!    DO iturb = n1,nturbines
       
!       i = turbines(id,iturb)%i
!       j = turbines(id,iturb)%j
      
!       IF(i.lt.mini)mini=i   ! 
!       IF(i.gt.maxi)maxi=i   ! save the min and max coordinates
!       IF(j.lt.minj)minj=j   ! of the turbines
!       IF(j.gt.maxj)maxj=j   !
       
!    END DO

!    !?? What happens if there is a domain without turbines
!    ncells(id) = n0
    

!    !-- count the number of cells containing turbines -----------------------
!    DO i=mini,maxi,n1
!       DO j=minj,maxj,n1
          
!          find     = .false.
!          iturb = n1
          
!          DO WHILE(.not.find.AND.iturb.le.nturbines)
             
!             ii = turbines(id,iturb)%i 
!             jj = turbines(id,iturb)%j
             
!             IF(ii.eq.i.and.jj.eq.j)THEN
!                ncells = ncells + n1
!                find   = .true.
!             END IF
             
!             iturb = iturb + n1
             
!          END DO
          
!       END DO
!    END DO


!    !Problem!!!!!!!!! ncells is domain dependent. Therefore we have to make 
!    !                 it nturbines which is the max possible amount of cells
!    !------------------------------------------------------------------------
!    !------  ALLOCATE cells and dummy cnt -----------------------------------
!    IF(.not.ALLOCATED(cells  ))ALLOCATE(  cells(max_domains,ncells))
!    IF(.not.ALLOCATED(TPC_CNT))ALLOCATE(TPC_CNT(max_domains,ncells))
!    !------------------------------------------------------------------------
!    !------------------------------------------------------------------------

!    indo = -1        ; indi =  1   ! N-S (-j) .. W-E ( i)
!    so   = maxj      ; fo   = minj
!    si   = mini      ; fi   = maxi

!    n  = n1 !1st cells index
!    nn = n1 !2nd cells index
!    !------------------------------------------------------------------------
!    !- Find the number of turbines that are in each cell                    -
!    !- Allocate component of cells%...                                      -
!    !- give the turbine numbers to the cells                                -
!    DO o=so,fo,indo    !o=outer
!       DO i=si,fi,indi !i=inner
!          !------------------------------------------------------------------
!          cnt = n0 ! cnt number of turbines in a cell
!          DO iturb = n1,nturbines
             
!             ii = turbines(id,iturb)%i 
!             jj = turbines(id,iturb)%j
             
!             !found a turbine in i,j:
!             IF(ii.EQ.i.AND.jj.EQ.o)cnt = cnt + n1

!             IF(iturb.EQ.nturbines.and.cnt.gt.n0)THEN
                
!                !------  ALLOCATE -------------------------------------------
!                IF(.not.ALLOCATED(cells(n)%turb_nr    ))&
!                     & ALLOCATE  (cells(n)%turb_nr(cnt))
!                !------------------------------------------------------------
!                cells(n)%turb_p_cell = cnt
!                TPC_CNT(n)           = cnt

!                n = n + n1
                
!             END IF
             
!          END DO

!          !------------------------------------------------------------------
!          cnt = n1 ! index
!          DO iturb = n1,nturbines
             
!             ii = turbines(id,iturb)%i 
!             jj = turbines(id,iturb)%j

!             IF(ii.EQ.i.AND.jj.EQ.o)THEN

!                IF(cnt.eq.n1)THEN
!                   cells(nn)%i = ii
!                   cells(nn)%j = jj
!                END IF

!                hcnt = n1
!                do k=n1,nturbines
!                   if(turbines(id,k)%i.eq.ii.and.turbines(id,k)%j.eq.jj)then
!                      if(hcnt.eq.cnt)cells(nn)%turb_nr(cnt) = turbines(id,k)%turb_nr
!                      hcnt = hcnt + n1
!                   end if
!                end do

!                cnt = cnt + n1
!             END IF

!             !found at least a turbine, then 
!             !we have to go to the next cell index
!             IF(cnt.GT.n1.AND.iturb.EQ.nturbines)&
!                  nn = nn + n1
!          END DO
!          !------------------------------------------------------------------
          
!       END DO
!    END DO


!  END SUBROUTINE PREP_CELL
!!! #########################################################################
!!! #########################################################################



  







!!! #########################################################################
!!! #########################################################################
  SUBROUTINE ABORT(text)
    
    IMPLICIT NONE
    

    CHARACTER(len=*)        :: text
    
    write(*,'(A)')trim(text); RETURN

    
  END SUBROUTINE ABORT
!!! #########################################################################
!!! #########################################################################



!!! #########################################################################
!!! ######################### END LOCAL ROUTINES ############################
!!! #########################################################################

END MODULE module_wf_ewp

