#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Aim: plot SAR 10 m wind alongside simulated WRF wind speed
#

import xarray as xr 
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import numpy as np

def add_ws_coords(ds,geod03):
    ue = ds.U10*geod03.COSALPHA.isel({'Time':0}) - ds.V10*geod03.SINALPHA.isel({'Time':0})    # Rotate Coordinates to Earth grid
    ve = ds.V10*geod03.COSALPHA.isel({'Time':0}) + ds.U10*geod03.SINALPHA.isel({'Time':0})    
    ds['WS'] = np.sqrt(ue**2 + ve**2)
    # load geogrid
    ds['XLAT_M'] = geod03.XLAT_M.isel({'Time':0})
    ds['XLONG_M'] = geod03.XLONG_M.isel({'Time':0})
    ds = ds.set_coords(['XLAT_M','XLONG_M'])
    return ds

def make_subplot(ax,ds,title,xlim=[5.5,9],ylim=[53,55.5],x='XLONG_M',y='XLAT_M',right_labels=False,left_labels=False,lcolorbar=False):
    pcol = ds.plot.pcolormesh(ax=ax,transform=ccrs.PlateCarree(),x=x,y=y,vmin=3,vmax=12,add_colorbar=lcolorbar,ylim=ylim,xlim=xlim,extend='both')
    ax.set_title(title)        
    ax.coastlines()
    gl2 = ax.gridlines(crs=ccrs.PlateCarree(),draw_labels=True, x_inline=False, y_inline=False)
    gl2.top_labels = False
    gl2.left_labels = left_labels
    gl2.right_labels = right_labels
    # adjust spacing
    plt.subplots_adjust(wspace=0.025)
    # adjust aspect ratio
    ax.set_aspect('auto')  
    return pcol

##### 0. Determine Font sizes
SMALL_SIZE = 18
MEDIUM_SIZE = 22
BIGGER_SIZE = 30
legendsize  = 12

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=legendsize)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title  

########################

# load SAR data and combine
sar1 = xr.open_dataset('sar/S1A_ESA_2017_10_14_17_17_12_0561316632_7.37E_53.31N_VV_C11_GFS025CDF_wind_level2.nc')
sar2 = xr.open_dataset('sar/S1A_ESA_2017_10_14_17_17_37_0561316657_6.90E_54.80N_VV_C11_GFS025CDF_wind_level2.nc')

sar_comb =  xr.concat([sar1.sar_wind,sar2.sar_wind.isel({'y':slice(1,-1)})],dim="y")
sar_comb1 =  xr.concat([sar1.mask,sar2.mask.isel({'y':slice(1,-1)})],dim="y")
sar_comb2 = xr.merge([sar_comb,sar_comb1])

# load model data (note: this model data does not include static variables, like coordinates or time to reduce disk space)
rootroot = '/mnt/airsea/LLJ/LLJS_2017101406/'
fitch   = xr.open_dataset(rootroot + 'results_025_new_corr/wrfout_d03_20171014_170000.nc')
fitch_1 = xr.open_dataset(rootroot + 'results_1_new_corr/wrfout_d03_20171014_170000.nc')
ewp     = xr.open_dataset(rootroot + '/results_ewp_corr/wrfout_d03_20171014_170000.nc')

# open geogrid files to add coordinates
geod03 = xr.open_dataset(rootroot + '/geo_em.d03.nc')

ewp    = add_ws_coords(ewp,geod03)
fitch   = add_ws_coords(fitch,geod03)
fitch_1 = add_ws_coords(fitch_1,geod03)

# plot
plt.close('all')
fig,axes = plt.subplots(1,4,figsize=(16,5),sharex=True,sharey=True,subplot_kw={'projection':ccrs.PlateCarree()})

pcol = make_subplot(axes[0],sar_comb2.sar_wind.where((sar_comb2.mask == -1) &( sar_comb2.sar_wind >2.5)),'SAR',x='longitude',y='latitude',left_labels=True)  
_    = make_subplot(axes[1],fitch.WS.isel({'Time':1}),'Fitch-on-0.25')  
_    = make_subplot(axes[2],fitch_1.WS.isel({'Time':1}),'Fitch-on-1')  
_    = make_subplot(axes[3],ewp.WS.isel({'Time':1}),'EWP-on',right_labels=True)  

# add colorbar below all 4 subplots
fig.colorbar(pcol, ax=axes.ravel().tolist(), shrink=0.7,orientation="horizontal",extend='both',label= '$WS_{10}$ [m/s]'    , pad=0.15,anchor=(0.5,1))

# wind farms scatter to subplots
wf = np.genfromtxt( rootroot + 'all_farms_corr_2017_new_LLJ_corrected.real')
wf_color = 'C7'
wf_alpha = 0.3
wf_label = 'turbines'
for ax in axes:
    ax.scatter(wf[:,2],wf[:,1],1,color=wf_color,label=wf_label,alpha=wf_alpha)

plt.show()
plt.savefig('sar/comp_sar_fitch_ewp_corrected.png', bbox_inches="tight", dpi=300)